
//
// Planet.java:
// Classes for planetary bodies,
// with methods for calculating elliptical orbits.
// Also the list of known objects in the solar system.
//    Copyright 1996 by Akkana Peck.
// You may use or distribute this code under the terms of the GPL.
//

package Akk.Astro;

import java.util.Date;
import java.lang.Math;
import java.awt.Graphics;
import java.awt.Button;
import java.awt.Event;
import java.awt.Font;
import java.applet.Applet;

// class planet -- a planetary body in the solar system.
// Instantiated with orbital elements & epoch.
// Given the current time, knows how to calc. heliocentric coordinates.
class Planet
{
    // orbital elements (all angles are in radians):
    String epoch;	// the epoch of these elements
    double e;		// eccentricity
			// a = q * (1 + e) / (1 - e*e);
    double node;	// longitude of asending node
    double w;		// argument of perihelion
    double i;		// inclination
    double epsilon;	// longitude of the planet at the epoch
    StarDate Tepoch;

    // quantities derived from the above, useful in calculations:
    double a;		// semi-major axis (AU)
    double wbar;
    double period;	// orbital period in earth years
    //double Tp;	// time of perihelion (tau), in decimal years
                        // relative to the epoch

    String name;

    static Planet earth
	= new Planet( "earth",
		     .016713,			// eccentricity
		     1.0,			// semi-major axis
		     0.,			// node
		     1.793647174,		// w = wbar
		     0.,			// inclination
		     1.734915012,		// epsilon
		     "1990 January 0" );	// epoch

    static Planet knownObjects[] = {
	// Planetary elements from Duffett-Smith:
	new Planet("Mercury",
		   0.205633,			// eccentricity
		   0.387099,			// SM axis
		   42.212740 * Math.PI/180,	// node
		   35.087093 * Math.PI/180,	// w = wbar - node
		   7.004540 * Math.PI/180,	// inclination
		   60.750646 * Math.PI/180,	// epsilon
		   "1990 January 0" ),		// epoch
	new Planet("Venus",
		   .006778,			// eccentricity
		   .723332,			// SM axis
		   76.589820 * Math.PI/180,	// node
		   54.840416 * Math.PI/180,	// w = wbar - node
		   3.394535 * Math.PI/180,	// inclination
		   88.455855 * Math.PI/180,	// epsilon
		   "1990 January 0" ),		// epoch
	new Planet("Mars",
		   .093396,			// eccentricity
		   1.523688,			// SM axis
		   49.480308 * Math.PI/180,	// node
		   286.394631 * Math.PI/180,	// w = wbar - node
		   1.849736 * Math.PI/180,	// inclination
		   240.739474 * Math.PI/180,	// epsilon
		   "1990 January 0" ),	// epoch
	new Planet("Jupiter",
		   0.048482,			// eccentricity
		   5.202561,			// SM axis
		   1.751492743,			// node
		   -1.498930562,		// w = wbar - node
		   0.022752339,			// inclination
		   1.581934756,			// epsilon
		   "1990 January 0" ),		// epoch
	new Planet("Saturn",
		   .055581,			// eccentricity
		   9.554747,			// SM axis
		   113.576139 * Math.PI/180,	// node
		   -20.714732 * Math.PI/180,	// w = wbar - node
		   2.488980 * Math.PI/180,	// inclination
		   287.690033 * Math.PI/180,	// epsilon
		   "1990 January 0" ),	// epoch
	new Planet("Uranus",
		   .046321,			// eccentricity
		   19.21814,			// SM axis
		   73.926961 * Math.PI/180,	// node
		   98.957872 * Math.PI/180,	// w = wbar - node
		   .773059 * Math.PI/180,	// inclination
		   271.063148 * Math.PI/180,	// epsilon
		   "1990 January 0" ),		// epoch
	new Planet("Neptune",
		   .009003,			// eccentricity
		   30.109570,			// SM axis
		   131.670599 * Math.PI/180,	// node
		   -83.660841 * Math.PI/180,	// w = wbar - node
		   1.770646 * Math.PI/180,	// inclination
		   282.249556 * Math.PI/180,	// epsilon
		   "1990 January 0" ),		// epoch
	// Elements for Pluto, also from Duffett-Smith, are the
	// osculating elements for epoch 1988 January 0.0
	new Planet("Pluto",
		   .24624,			// eccentricity
		   39.3414,			// SM axis
		   110.144 * Math.PI/180,	// node
		   133.989 * Math.PI/180,	// w = wbar - node
		   17.1420 * Math.PI/180,	// inclination
		   221.4127 * Math.PI/180,	// epsilon
		   "1988 January 0" ),		// epoch
	// H-B elements from http://encke.jpl.nasa.gov/eph/95O1.2000_yeomans
	// Most recently, from "Comet Comments" posting.
	new Planet("Hale-Bopp",
		   0.995103273,			// eccentricity
		   186.6793275,		// a = q * (1 + e) / (1 - e*e);
		   282.4708087*Math.PI/180,	// node
		   130.5909555*Math.PI/180,	// w
		   89.42965142*Math.PI/180,	// inclination
		   "1997 Apr  1.13843",		// Tp
		   "1997 May 3.00000" ),		// epoch
	// Kopff elements from http://encke.jpl.nasa.gov/eph/22P.2000_yeomans:
	new Planet("Kopff",
		   0.543940053,			// eccentricity
		   3.464496901,			// SM axis
		   120.9108664*Math.PI/180,	// node = LOMEGA
		   162.8351167*Math.PI/180,	// w = SOMEGA
		   4.7208259*Math.PI/180,	// inclination
		   "1996 Jul 2.1939239",	// Tp
		   "1996 Jul 1" ),		// epoch
	// Brewington elements from http://www.execpc.com/~skymap
	// Brewington uses parabolic orbital elements.
	new Planet("Brewington",
		   .925734,			// q
		   234.9092*Math.PI/180,	// node (= LOMEGA)
		   43.9782*Math.PI/180,		// w (= SOMEGA)
		   52.1415*Math.PI/180,		// i
		   "1996 Aug 3.43037" ),	// time of perihelion
						// (JD 2450298.92997)
	// Comet Tabur is also parabolic: elements off Don Machholz'
	// "Comet Comments" posting
	new Planet("Tabur",
		   0.84001480,			// q
		   31.41231*Math.PI/180,	// node
		   57.37495*Math.PI/180,	// w
		   73.36167*Math.PI/180,	// i
		   "1996 Nov 3.50419" ),	// time of perihelion
	// Machholz 1 elements from "Comet Comments"
	new Planet("Machholz 1",
		   0.9586366,			// eccentricity
		   3.015172792,  		// a = q * (1 + e) / (1 - e*e);
		   94.53200*Math.PI/180,	// node
		   14.58608*Math.PI/180,	// w
		   60.07415*Math.PI/180,	// inclination
		   "1996 Oct 15.06962",		// Tp
		   "2000 Jan 1" ),		// epoch
	/* Kohler was only here for testing vs. Duffett-Smith:
	new Planet("Kohler",
		   .990662,			// q
		   181.8175*Math.PI/180,	// node
		   163.4799*Math.PI/180,	// w
		   48.7196*Math.PI/180,		// i
		   "1977 Nov 10.5659" ),	// time of peri
	 */
    };

    // Constructor -- takes orbital elements plus epoch.
    // Expects angles already in radians.
    public Planet(String s,
		  double e1, double a1, double node1, double w1, double i1,
		  double ep1, String epoch)
    {
	// Copy non-date args:
	name = s;
	e = e1;
	a = a1;
	node = node1;
	i = i1;
	w = w1;
	epsilon=ep1;

	// Use Kepler's law to calculate the period in decimal earth years:
	period = Math.sqrt(a*a*a);

	// convert times from strings to StarDate objects:
	Tepoch = new StarDate(epoch);

	// calculate wbar from w and node:
	wbar = w + node;
    }

    // Alternate constructor -- takes Tp instead of epsilon
    public Planet(String s,
		  double e1, double a1, double node1, double w1, double i1,
		  String Tp1, String epoch)
    {
	// Copy non-date args:
	name = s;
	e = e1;
	a = a1;
	node = node1;
	i = i1;
	w = w1;

	// convert times from strings to StarDate objects:
	Tepoch = new StarDate(epoch);

	// Calculate epsilon from Tp1:
	StarDate tp = new StarDate(Tp1);
	double dt = tp.decimalYears() - Tepoch.decimalYears();
	period = Math.sqrt(a*a*a);
	double n = 2. * Math.PI / period;
	epsilon = w + node - n * dt;

	// calculate wbar from w and node:
	wbar = w + node;
    }

    // Third constructor, for parabolic orbits: can't specify a, only q.
    // e is assumed to be 1.0 in this ctor
    public Planet(String s,
		  double q1, double node1, double w1, double i1,
		  String tper)
    {
	e = 1.0;

	// store q in the variable a, since a is meaningless:
	a = q1;
	// store time of perihelion in the variable Tepoch
	Tepoch = new StarDate(tper);

	// store the other elements normally:
	node = node1;
	w = w1;
	i = i1;
	name = s;

	// calculate wbar from w and node:
	wbar = w + node;
    }

    // Calculate the right ascension and declination at a given time
    // for the object with these orbital elements.
    // Equations derived from Peter Duffett-Smith,
    // PRACTICAL ASTRONOMY WITH YOUR CALCULATOR,
    // and W. M. Smart, TEXTBOOK ON SPHERICAL ASTRONOMY.
    public HelioCoords calcHeliocentric(StarDate Tnow)
    {
	// First, print out what we think the current orbital elements are:
//	System.out.println("Trying to calc coords for " + name);
//	System.out.println("a = " + new Double(a).toString());
//	System.out.println("e = " + new Double(e).toString());
//	System.out.println("i = " + new Double(i*180/Math.PI).toString());
//	System.out.println("w = " + new Double(w*180/Math.PI).toString());
//	System.out.println("wbar = " + new Double(wbar*180/Math.PI).toString());
//	System.out.println("node = " + new Double(node*180/Math.PI).toString());
//	System.out.println("epsilon = " + new Double(epsilon*180/Math.PI).toString());
//	System.out.println("period = " + new Double(period).toString());

	double v=0, r=0, l=0;

	//
	// Getting the true anomaly differs depending on whether the
	// orbit is elliptical or parabolic:
	//
	if (e < 1.0)		// normal elliptical orbit
	{

	    // n is the average angular velocity:
	    double n = 2. * Math.PI / period;
//	    System.out.println("n = " + new Double(n*180/Math.PI).toString());

	    // N is n times decimal years since epoch, basically the current
	    // position past perihelion based on average angular velocity
	   double D=Tnow.decimalYears() - Tepoch.decimalYears();
//	    System.out.println("D = " + new Double(D).toString() + " years");
//	    System.out.println("D = " + new Double(D*365.242191).toString() + " days");

	    double N = n * (Tnow.decimalYears() - Tepoch.decimalYears());
	    // and make sure it's between 0 and 2*Math.PI:
	    while (N > 2.*Math.PI)
		N -= 2*Math.PI;
	    while (N < 0.)
		N += 2*Math.PI;
//	    System.out.println("N = " + new Double(N*180/Math.PI).toString());


	    // Find the mean anomaly
	    double M = N + epsilon - wbar;
//	    System.out.println("M = " + new Double(M*180/Math.PI).toString());

//	    System.out.println("Trying to find true anomaly ...");
	    // Find the true anomaly v and heliocentric longitude l
	    // (method varies depending on eccentricity):
	    if (e < .1) {
		// For low eccentricity, the "equation of the center"
		// is a good and fast approximation:

		l = (N + 2. * e * Math.sin(M) + epsilon);
		// and make sure it's between 0 and 2*Math.PI:
		while (l > 2.*Math.PI)
		    l -= 2*Math.PI;
		while (l < 0.)
		    l += 2*Math.PI;

		// and find the true anomaly:
		v = l - wbar;

	    } else {
		// For higher eccentricity (e.g. comets),
		// we have to iterate and solve Kepler's equation
		// using a delta-epsilon algorithm.  (Note, this
		// ep has nothing to do with the epsilon which denotes
		// the longitude of the planet at the epoch.)
		double ep=.00001;
		double del=0;
		double E=M;
		do {
		    E -= del / (1. - e * Math.cos(E));
		    del = E - e * Math.sin(E) - M;
		} while (Math.abs(del) > ep);
		// Now E is the eccentric anomaly

		v = 2. * Math.atan( Math.sqrt((1 + e) / (1 - e))
				   * Math.tan(E/2) );
	    }
	    r = a * (1 - e * e) / (1 + e * Math.cos(v));
	}	// end of elliptical orbit -> true anomaly
	else if (e == 1.0)	// parabolic orbit
	{
	    // perihelion distance is stored in the variable "a"
	    double q = a;

	    // find d = number of days since perihelion:
	    double d = (Tnow.getTime() - Tepoch.getTime()) / (24.*60*60*1000);
//	    System.out.println("days since peri: " + new Double(d).toString());

	    double W = 0.0364911624 * d / q / Math.sqrt(q);

//	    System.out.println("Trying to find hyperbolic true anomaly:");
//	    System.out.println("W = " + new Double(W).toString());
	    // Use an iterative method to solve the equation
	    // s^3 * 3s - W = 0.
	    double s = W/3;
	    while (true) {
//		System.out.println("s = " + new Double(s).toString());
		double delta = s*s*s + 3*s - W;
		if (Math.abs(delta) < .000001)
		    break;
		s = (2*s*s*s + W) / (3 * (s*s + 1));
	    }

	    // Now calculate the true anomaly:
	    v = 2 * Math.atan(s);
	    r = q * (1 + s*s);
	}
	//else throw HyperbolicException;

	// Calc longitude from true anomaly:
	l = v + wbar;

//	System.out.println("l = " + new Double(l*180/Math.PI).toString());
//	System.out.println("v = " + new Double(v*180/Math.PI).toString());

	// Now we have the true anomaly v, heliocentric distance r,
	// and longitude l.
	// Find heliocentric coordinates to return:
	double x = Math.cos(l - node);
	double y = Math.sin(l - node) * Math.cos(i);
//	System.out.println("x = " + new Double(x).toString());
//	System.out.println("y = " + new Double(y).toString());
//	System.out.println("r = " + new Double(r).toString());
	return new HelioCoords(r, x, y, l);
    }

    //
    // locate(): Calculate geocentric coordinates of a Planet object.
    // Equations derived from Peter Duffett-Smith,
    // PRACTICAL ASTRONOMY WITH YOUR CALCULATOR,
    // and W. M. Smart, TEXTBOOK ON SPHERICAL ASTRONOMY.
    //
    AstroCoords locate(StarDate now, HelioCoords earthc)
    {
	HelioCoords planetc = calcHeliocentric(now);

	// Project radius of our position onto the ecliptic:
	double psi = Math.asin(Math.sin(planetc.l - node)
			       * Math.sin(i));
//	System.out.println("psi = Math.asin( Math.sin("
//			   + new Double(planetc.l*180/Math.PI).toString()
//			   + " - "
//			   + new Double(node*180/Math.PI).toString()
//			   + ") * Math.sin("
//			   + new Double(i*180/Math.PI).toString()
//			   + "))");
//	System.out.println("psi = " + new Double(psi*180/Math.PI).toString());

	double rr = planetc.r * Math.cos(psi);
	if (rr < 0) rr = -rr;

	// Find the arc tangent of y/x, and correct for the right quadrant
	// (want atn to be between -Math.PI and Math.PI).
	// It would be nice if atan2 did this, but in practice, atan2
	// gives a completely wrong answer!
	// This seems to be because the doc is wrong -- the doc says
	// atan(a,b) gives atan(b/a), but actually it gives atan(a/b)!
	double ll = Math.atan2(planetc.y, planetc.x) + node;

	double A = Math.atan( (rr * Math.sin(earthc.l - ll)) /
			     (earthc.r - rr * Math.cos(earthc.l - ll)) );
//	System.out.println("A = " + new Double(A*180/Math.PI).toString());

	// Figure out which equation to use, depending on whether the
	// comet is closer to or farther from the sun than we are:
	double lambda;		// ecliptic longitude
	if (rr < earthc.r)
	    lambda = Math.PI + earthc.l
	            + Math.atan( (rr * Math.sin(earthc.l - ll)) /
				(earthc.r - rr * Math.cos(earthc.l - ll)) );

	else
	    lambda = ll
	            + Math.atan( (earthc.r * Math.sin(ll - earthc.l)) /
				(rr - earthc.r * Math.cos(ll - earthc.l)) );

	if (lambda < 0)
	    lambda += 2 * Math.PI;

	// ecliptic longitude:
	double beta = Math.atan( (rr * Math.tan(psi) * Math.sin(lambda - ll)) /
				(earthc.r * Math.sin(ll - earthc.l)) );


//	System.out.println("lambda = " + new Double(lambda*180/Math.PI).toString());
//	System.out.println("beta = " + new Double(beta*180/Math.PI).toString());

	// Finally, calculate RA and Dec:
	double ecliptic = 23.4419 * Math.PI / 180.;
	double RA = Math.atan2( (Math.sin(lambda) * Math.cos(ecliptic)
			      - Math.tan(beta) * Math.sin(ecliptic)),
			   Math.cos(lambda) );
	if (RA < 0)
	    RA += 2. * Math.PI;

	double dec = Math.asin(Math.sin(beta) * Math.cos(ecliptic)
			    + Math.cos(beta) * Math.sin(ecliptic)
			      * Math.sin(lambda));

	double dist = Math.sqrt(earthc.r * earthc.r + planetc.r * planetc.r
			     - (2 * earthc.r * planetc.r
				* Math.cos(planetc.l - earthc.l)));

	AstroCoords ret = new AstroCoords(RA, dec, dist);

	System.out.println("Found " + name + ": " + ret.toString());
	return ret;
    }
}

