
//
// DrawJupiter.java -- display positions of Jupiter's moons.
//	Copyright 1996 by Akkana Peck.
//

package Akk.Astro;

import java.awt.*;
import java.applet.*;
import java.util.Date;

public class DrawJupiter extends Applet
{
    Jupiter jupiter;

    // Size constants:
    final static int jsize = 30;	// planet's radius in pixels
    final static int moonsize = 12;	// moons' size in pixels
    final static int redSpotWidth = 21;
    final static int redSpotHeight = 15;

     // Copies of the UI components we maintain:
    Canvas drawingArea;
    TextField dateField;
    Choice invertMenu;
    Panel controlPanel;

    boolean monochromeP;	// will it work through a red filter?
    Color planetColor;
    Color moonColor;
    Color shadowColor;
    Color shadowLablColor;
    Color lablColor;
    Color bandColor;
    Color spotColor;

    int spotLongitude;

    public String getAppletInfo() {
	return "Jovian Moon Locator.\nCopyright 1996 by Akkana Peck";
    }

    public String[][] getParameterInfo() {
	String[][] info = {
	    { "time", "string",
	      "Time of chart (optional, defaults to current)" },
            { "monochrome", "boolean",
              "Don't use colors (good for use with red filters)" },
            { "spotLongitude", "Integer",
              "Longitude of the GRS in System II" },
   	};
	return info;
    }

    public void init()
    {
	planetColor = null;

	StarDate now = null;
    	String datestring = getParameter("time");
	if (datestring == null)
            try {
                now = new StarDate();
                datestring = now.toString();
            }
	    catch (IllegalArgumentException e) {
		datestring = "Internal java error: can't get current date!";
            }
	else
            try {
                now = new StarDate(datestring);
            }
	    catch (IllegalArgumentException e) {
		datestring = "Can't parse date: " + datestring;
            }

        String s = getParameter("monochrome");
        if (s != null)
            monochromeP = Boolean.valueOf(s).booleanValue();
        else
            monochromeP = false;

        s = getParameter("spotLongitude");
        if (s != null)
            spotLongitude = Integer.valueOf(s).intValue();
        else
            spotLongitude = 66;

	//
	// Set some layout info:
	//
	setLayout(new BorderLayout(0, 1));

	//
	// Create the star chart object
	//
	drawingArea = new Canvas();
	add("North", drawingArea);

	dateField = new TextField(datestring, 12);
	add("South", dateField);
    }

    public boolean handleEvent(Event event)
    {
	if (event.id == Event.ACTION_EVENT && event.target == dateField )
	{
	    try {
		StarDate newdate = new StarDate(dateField.getText());
		// Might throw IllegalArgumentException if user
		// typed in the wrong format.
                repaint();
	    }
	    catch (IllegalArgumentException e) {
                String bad = dateField.getText();
		dateField.setText("Can't parse date: " + bad);
	    }
	    finally {
		return true;
	    }
	}

	return super.handleEvent(event);
    }

    public void paint(Graphics g)
    {
        Dimension drawsize = new Dimension(size().width,
                                           size().height -
					     dateField.size().height);
    	int width = drawsize.width;
        int height = drawsize.height;
    	int xcenter = width/2;
    	int ycenter = height/2;

        try {
            StarDate date = new StarDate(dateField.getText());

            if (jupiter == null)
                jupiter = new Jupiter(date);
            else
                jupiter.setDate(date);

	// Do some color manipulation:
            if (planetColor == null)
            {
		if (monochromeP)
		{
		    drawingArea.setBackground(Color.black);
		    drawingArea.setForeground(Color.white);
		    planetColor = Color.white;
		    moonColor = drawingArea.getForeground();
		    shadowColor = Color.black;
		    shadowLablColor = Color.lightGray;
		    lablColor = Color.white;
		    bandColor = Color.lightGray;
		    spotColor = Color.lightGray;
		}
		else
		{
		    drawingArea.setBackground(Color.black);
		    drawingArea.setForeground(Color.white);
		    planetColor = Color.yellow;
		    moonColor = drawingArea.getForeground();
		    shadowColor = Color.black;
		    shadowLablColor = Color.cyan;
		    lablColor = Color.white;
		    bandColor = Color.orange;
		    spotColor = Color.red;
		}
            }

            g.setColor(drawingArea.getBackground());
            g.fillRect(0, 0, width, height);

            // draw the planet itself:
            g.setColor(planetColor);
            g.fillOval(xcenter-jsize, ycenter-jsize, jsize*2, jsize*2);

            XYCoord coord;

            // Draw a couple of fake bands:
            g.setColor(bandColor);
            for (int i=0; i < 4; ++i) {
                g.drawLine(xcenter-jsize+i, ycenter-i-2,
                           xcenter+jsize-i, ycenter-i-2);
                g.drawLine(xcenter-jsize+i, ycenter+i+2,
                           xcenter+jsize-i, ycenter+i+2);
            }

            // Find the red spot:
            coord = jupiter.getRedSpotXY(spotLongitude);
            if (coord.x != Double.POSITIVE_INFINITY
                && coord.y != Double.POSITIVE_INFINITY)
            {
                g.setColor(spotColor);
                // We only want the spot to draw to the edge of the disk,
                // so adjust the width if necessary:
                // (approx) width/2 of the planet (pixels) at the spot's latitude:
                int j2width = (int)(jsize - Math.abs(coord.y * jsize)/3);
                int x = xcenter + (int)(coord.x * jsize);
                int y = ycenter + (int)(coord.y * jsize) - redSpotHeight/2;
                int spotwidth = redSpotWidth;
                if (x + redSpotWidth > xcenter + j2width) {
                    spotwidth = xcenter + j2width - x;
                    x = xcenter + j2width - spotwidth;
                } else if (x - redSpotWidth/2 < xcenter - j2width) {
                    spotwidth = x + redSpotWidth/2 - xcenter + j2width;
                    x = xcenter - j2width;
                }
                else {
                    x -= redSpotWidth/2;
                }

                g.fillOval(x, y, spotwidth, redSpotHeight);
            }

            // draw each of the moons:
            int whichmoon;
            for (whichmoon=0; whichmoon < jupiter.getNumMoons(); ++whichmoon)
            {
                // See if this moon is casting a shadow:
                coord = jupiter.getMoonShadowXY(whichmoon);
                if (coord.x != Double.POSITIVE_INFINITY
                    && coord.y != Double.POSITIVE_INFINITY)
                {
                    int x = xcenter + (int)(coord.x * jsize) - moonsize/2;
                    int y = ycenter - (int)(coord.y * jsize) - moonsize/2;
                    g.setColor(shadowColor);
                    g.fillOval(x, y, moonsize, moonsize);
                    g.setColor(shadowLablColor);
                    g.drawString(jupiter.getMoonLabel(whichmoon),
                                 x /*+ moonsize*/, ycenter - 5*moonsize);
                }

                coord = jupiter.getMoonXY(whichmoon);
                if ((coord.x != Double.POSITIVE_INFINITY)
                    && (coord.y != Double.POSITIVE_INFINITY))
                {
                    int x = xcenter + (int)(coord.x * jsize) - moonsize/2;
                    int y = ycenter - (int)(coord.y * jsize) - moonsize/2;

                    g.setColor(drawingArea.getBackground());
                    g.drawOval(x-1, y-1, moonsize+2, moonsize+2);
                    g.setColor(moonColor);
                    g.fillOval(x, y, moonsize, moonsize);

                    g.setColor(lablColor);
                    g.drawString(jupiter.getMoonLabel(whichmoon),
                                 x /*+ moonsize*/, ycenter - 3*moonsize);
                }
            } // end loop over moons
        } // end "try" clause for successful paint

        catch (IllegalArgumentException e) {
            g.setColor(Color.black);
            g.fillRect(0, 0, width, height);
            g.setColor(Color.yellow);
            g.drawString("Error: can't parse date " + dateField.getText(),
                         xcenter, ycenter);
        }
    } // end paint()
}

